package api

import (
	"context"
	"encoding/json"
	"fmt"
	"log"
	"net/http"
	"strconv"
	"strings"
	"time"

	"golang.org/x/oauth2"

	"github.com/gorilla/mux"

	"bravo.sau.kau.se/rethink/planera-mera/internal/app/planeramera"
	"bravo.sau.kau.se/rethink/planera-mera/internal/app/planeramera/canvas"
	"bravo.sau.kau.se/rethink/planera-mera/internal/app/planeramera/datastore"
	"bravo.sau.kau.se/rethink/planera-mera/internal/app/planeramera/lti"
	"bravo.sau.kau.se/rethink/planera-mera/internal/app/planeramera/session"
)

// API represents the API
type API struct {
	sessionStore     *session.MemorySessionStore
	dataStore        *datastore.Datastore
	oauth2Conf       *oauth2.Config
	canvasBaseURL    string
	tokenStoragePool planeramera.TokenStoragePool
}

// NewHandler returns a new API handler
func NewHandler(s *session.MemorySessionStore,
	ds *datastore.Datastore,
	oauth2Conf *oauth2.Config,
	canvasurl string) *API {

	return &API{
		sessionStore:     s,
		dataStore:        ds,
		canvasBaseURL:    canvasurl,
		oauth2Conf:       oauth2Conf,
		tokenStoragePool: planeramera.NewTokenStoragePool(),
	}
}

// ContextKey is key for context values
type ContextKey string

// LaunchInfo represents LTI launch info from Canvas
type LaunchInfo struct {
	CourseCode   string `json:"course_code"`
	CourseName   string `json:"course_name"`
	IsInstructor bool   `json:"is_instructor"`
}

// Module represents an canvas module
type Module struct {
	ID          int          `json:"id,omitempty"`
	Name        string       `json:"name,omitempty"`
	Items       []ModuleItem `json:"items,omitempty"`
	Published   bool         `json:"published,omitempty"`
	Instruction string       `json:"instruction,omitempty"`
}

// ModuleItem represents an item in module
type ModuleItem struct {
	ID           int          `json:"id,omitempty"`
	Title        string       `json:"title,omitempty"`
	Type         string       `json:"type,omitempty"`
	URL          string       `json:"url,omitempty"`
	Indent       int          `json:"indent,omitempty"`
	Published    bool         `json:"published,omitempty"`
	Activity     Activity     `json:"activity,omitempty"`
	ActivityUser ActivityUser `json:"activity_user,omitempty"`
}

// Category represents a category that can be set for a planned activity
type Category struct {
	ID   int    `json:"id,omitempty"`
	Name string `json:"name,omitempty"`
}

// Activity represents a planned activity for a module item
type Activity struct {
	StartDate  *time.Time `json:"startdate,omitempty"`
	StopDate   *time.Time `json:"stopdate,omitempty"`
	Estimated  *float32   `json:"estimated,omitempty"`
	Updated    time.Time  `json:"updated,omitempty"`
	Visible    bool       `json:"visible,omitempty"`
	Categories []Category `json:"categories,omitempty"`
}

// StartStop represents a start- and stopdate
type StartStop struct {
	Start *time.Time `json:"start,omitempty"`
	Stop  *time.Time `json:"stop,omitempty"`
}

// FloatValue represents an integer value
type FloatValue struct {
	Value *float32 `json:"value,omitempty"`
}

// StringValue represents a string value
type StringValue struct {
	Value *string `json:"value,omitempty"`
}

// CalendarEvent represents an event to add to calendar
type CalendarEvent struct {
	Title       string `json:"title,omitempty"`
	Description string `json:"description,omitempty"`
	StartDate   string `json:"startdate,omitempty"`
	StopDate    string `json:"stopdate,omitempty"`
}

// ActivityUser represent a planned activity for a module item for a specific user
type ActivityUser struct {
	Estimated *float32   `json:"estimated,omitempty"`
	Imported  *time.Time `json:"imported,omitempty"`
}

// Planning represents the planning for the current course
type Planning struct {
	Activated *time.Time `json:"activated,omitempty"`
}

// assertUserIsInstructor checks that the current user is an instructor.
// Used in PUT methods.
func assertUserIsInstructor(w http.ResponseWriter, r *http.Request) bool {
	userID := r.Context().Value(ContextKey("userID")).(int)

	if !userIsInstructor(r) {
		log.Printf("API.assertUserIsInstructor: user %d is not an instructor", userID)
		http.Error(w, http.StatusText(http.StatusUnauthorized), http.StatusUnauthorized)
		return false
	}

	return true
}

// checks if current user is an instructor
func userIsInstructor(r *http.Request) bool {
	roles := r.Context().Value(ContextKey("roles")).(string)
	return strings.Contains(strings.ToLower(roles), "instructor")
}

// tokenFromSession creates an oauth2.Token from values in session
func tokenFromSession(s *session.MemorySession) (oauth2.Token, error) {
	t := oauth2.Token{
		AccessToken:  s.Get("AccessToken"),
		RefreshToken: s.Get("RefreshToken"),
		TokenType:    s.Get("TokenType"),
	}
	if t.AccessToken == "" || t.RefreshToken == "" || t.TokenType == "" {
		return t, fmt.Errorf("missing token data in session")
	}
	if err := t.Expiry.UnmarshalText([]byte(s.Get("Expiry"))); err != nil {
		return t, err
	}
	return t, nil
}

// SessionMW checks if API call is authorized
func (api *API) SessionMW(next http.Handler) http.Handler {
	return http.HandlerFunc(func(w http.ResponseWriter, r *http.Request) {
		auth := r.Header.Get("Authorization")
		if auth == "" {
			log.Printf("Missing Authorization header. Denying access.")
			http.Error(w, "authorization failed", http.StatusUnauthorized)
			return
		}
		parts := strings.SplitN(auth, " ", 2)
		if len(parts) != 2 || parts[0] != "PM" {
			log.Printf("Authorization header is wrong type: %v", auth)
			http.Error(w, "wrong authorization type", http.StatusUnauthorized)
			return
		}

		session := api.sessionStore.Session(parts[1])
		token, err := tokenFromSession(session)
		if err != nil {
			log.Printf("%v", err)
			http.Error(w, "invalid session", http.StatusUnauthorized)
			return
		}

		tokenStorage := api.tokenStoragePool.TokenStorageByTokenID(session.Get("tokenUserID"))
		if tokenStorage == nil {
			tokenStorage = &planeramera.TokenStorage{
				New: api.oauth2Conf.TokenSource(context.Background(), &token),
				T:   &token,
				F: func(t *oauth2.Token) error {
					lti.TokenToSession(session, t.AccessToken, t.RefreshToken, t.TokenType, t.Expiry)
					return api.dataStore.TokenSave(session.Get("tokenUserID"), t.AccessToken, t.RefreshToken, t.TokenType, t.Expiry)
				},
			}
			api.tokenStoragePool.AddTokenStorage(session.Get("tokenUserID"), tokenStorage)
		}

		client := oauth2.NewClient(r.Context(), tokenStorage)
		canvasClient := canvas.NewClient(api.canvasBaseURL, client)
		courseID, _ := strconv.Atoi(session.Get("courseID"))
		courseCode := session.Get("courseCode")
		courseName := session.Get("courseName")
		userID, _ := strconv.Atoi(session.Get("userID"))
		roles := session.Get("roles")
		ctx := context.WithValue(r.Context(), ContextKey("canvasClient"), canvasClient)
		ctx = context.WithValue(ctx, ContextKey("courseID"), courseID)
		ctx = context.WithValue(ctx, ContextKey("courseCode"), courseCode)
		ctx = context.WithValue(ctx, ContextKey("courseName"), courseName)
		ctx = context.WithValue(ctx, ContextKey("userID"), userID)
		ctx = context.WithValue(ctx, ContextKey("roles"), roles)

		next.ServeHTTP(w, r.WithContext(ctx))
	})
}

// HeaderSetterMW sets common response headers
func HeaderSetterMW(next http.Handler) http.Handler {
	return http.HandlerFunc(func(w http.ResponseWriter, r *http.Request) {
		w.Header().Set("Access-Control-Allow-Origin", "*")
		w.Header().Set("Cache-Control", "no-cache, no-store, must-revalidate")
		w.Header().Set("Pragma", "no-cache")
		w.Header().Set("Expires", "0")
		next.ServeHTTP(w, r)
	})
}

// writeJSON
func writeJSON(w http.ResponseWriter, r *http.Request, data interface{}) {
	jsondata, err := json.MarshalIndent(data, "", "  ")
	if err != nil {
		http.Error(w, err.Error(), http.StatusInternalServerError)
	}
	r.Header.Set("Content-Type", "application/json")
	fmt.Fprint(w, string(jsondata))
}

// SlowerMW slows down all requests
func SlowerMW(next http.Handler) http.Handler {
	return http.HandlerFunc(func(w http.ResponseWriter, r *http.Request) {
		time.Sleep(2 * time.Second)
		next.ServeHTTP(w, r)
	})
}

// RegisterRoutes register routes using the given router
func (api *API) RegisterRoutes(r *mux.Router) {
	r.Use(HeaderSetterMW)
	r.Use(api.SessionMW)
	//r.Use(SlowerMW)
	r.HandleFunc("/categories", api.Categories).Methods("GET")
	r.HandleFunc("/ltiinfo", api.LaunchInfo).Methods("GET")
	r.HandleFunc("/modules", api.Modules).Methods("GET")
	r.HandleFunc("/module/{id:[0-9]+}/instruction", api.ModuleInstructionPUT).Methods("PUT").Headers("Content-Type", "application/json")
	r.HandleFunc("/planning", api.Planning).Methods("GET")
	r.HandleFunc("/planning/active", api.PlanningActivePUT).Methods("PUT").Headers("Content-Type", "application/json")
	r.HandleFunc("/moduleitem/{id:[0-9]+}/activity/visible", api.ActivityVisiblePUT).Methods("PUT").Headers("Content-Type", "application/json")
	r.HandleFunc("/moduleitem/{id:[0-9]+}/activity/startstop", api.ActivityStartStopPUT).Methods("PUT").Headers("Content-Type", "application/json")
	r.HandleFunc("/moduleitem/{id:[0-9]+}/activity/estimated", api.ActivityEstimatedPUT).Methods("PUT").Headers("Content-Type", "application/json")
	r.HandleFunc("/moduleitem/{id:[0-9]+}/activity/categories", api.ActivityCategoriesPOST).Methods("POST").Headers("Content-Type", "application/json")
	r.HandleFunc("/moduleitem/{id:[0-9]+}/activity/category/{category_id:[0-9]+}", api.ActivityCategoryDELETE).Methods("DELETE")
	r.HandleFunc("/moduleitem/{id:[0-9]+}/activityUser/estimated", api.ActivityUserEstimatedPUT).Methods("PUT").Headers("Content-Type", "application/json")
	r.HandleFunc("/moduleitem/{id:[0-9]+}/calendarEvent", api.CalendarEventPOST).Methods("POST").Headers("Content-Type", "application/json")
}

// LaunchInfo returns LTI informations
func (api *API) LaunchInfo(w http.ResponseWriter, r *http.Request) {
	courseCode := r.Context().Value(ContextKey("courseCode")).(string)
	courseName := r.Context().Value(ContextKey("courseName")).(string)

	ret := LaunchInfo{
		CourseCode:   courseCode,
		CourseName:   courseName,
		IsInstructor: userIsInstructor(r),
	}

	writeJSON(w, r, ret)
}

// Modules returns all modules for current course
func (api *API) Modules(w http.ResponseWriter, r *http.Request) {
	var modules = []Module{}

	// get course id from sessions
	courseID, ok := r.Context().Value(ContextKey("courseID")).(int)
	if !ok {
		log.Print("API.Modules: invalid courseID")
		http.Error(w, http.StatusText(http.StatusInternalServerError), http.StatusInternalServerError)
		return
	}

	// students should only get modules if planning is activated
	p, err := api.dataStore.PlanningByCourseID(courseID)
	if err != nil && !userIsInstructor(r) && err != datastore.ErrNoResult {
		log.Printf("API.Modules PlanningByCourseID: %v", err)
		http.Error(w, http.StatusText(http.StatusInternalServerError), http.StatusInternalServerError)
		return
	}
	if !userIsInstructor(r) && p.Activated == nil {
		writeJSON(w, r, modules)
		return
	}

	// get modules and module items from canvas
	canvasClient, ok := r.Context().Value(ContextKey("canvasClient")).(*canvas.Client)
	if !ok {
		log.Print("API.Modules: no Canvas client in context")
		http.Error(w, http.StatusText(http.StatusInternalServerError), http.StatusInternalServerError)
		return
	}

	canvasModules, err := canvasClient.Modules(courseID)
	if err != nil {
		log.Printf("API.Modules Modules: %v", err)
		http.Error(w, http.StatusText(http.StatusInternalServerError), http.StatusInternalServerError)
		return
	}

	for _, canvasModule := range canvasModules {
		module := Module{
			ID:        canvasModule.ID,
			Name:      canvasModule.Name,
			Published: canvasModule.Published,
		}

		// get module instruction from database
		i, err := api.dataStore.InstructionByModuleID(module.ID)
		if err != nil {
			if err != datastore.ErrNoResult {
				log.Printf("API.Modules InstructionByModuleID: %v", err)
				http.Error(w, "failed to load instruction", http.StatusInternalServerError)
				return
			}
		}
		module.Instruction = i.Instruction

		// create array of module items
		for _, canvasItem := range canvasModule.Items {
			item := ModuleItem{
				ID:        canvasItem.ID,
				Title:     canvasItem.Title,
				Type:      canvasItem.Type,
				URL:       canvasItem.URL,
				Indent:    canvasItem.Indent,
				Published: canvasItem.Published,
			}

			// get activity planned for module item from database
			a, err := api.dataStore.ActivityByModuleItemID(item.ID)
			if err != nil {
				if err != datastore.ErrNoResult {
					log.Printf("API.Modules ActivityByModuleItemID : %v", err)
					http.Error(w, "failed to load activities", http.StatusInternalServerError)
					return
				}
			}

			item.Activity = Activity{
				Visible:   a.Visible,
				StartDate: a.StartDate,
				StopDate:  a.StopDate,
				Estimated: a.Estimated,
				Updated:   a.Updated,
			}

			for _, c := range a.Categories {
				var category = Category{
					ID:   c.ID,
					Name: c.Name,
				}
				item.Activity.Categories = append(item.Activity.Categories, category)
			}

			// get userid from session
			userID, ok := r.Context().Value(ContextKey("userID")).(int)
			if !ok {
				log.Print("API.Modules: invalid userID")
				http.Error(w, http.StatusText(http.StatusInternalServerError), http.StatusInternalServerError)
				return
			}

			// get activity planned for module item by user from database
			au, err := api.dataStore.ActivityUserByModuleItemID(item.ID, userID)
			if err != nil {
				if err != datastore.ErrNoResult {
					log.Printf("API.Modules ActivityUserByModuleItemID : %v", err)
					http.Error(w, "failed to load activities for user", http.StatusInternalServerError)
					return
				}
			} else {
				item.ActivityUser = ActivityUser{
					Estimated: au.Estimated,
					Imported:  au.Imported,
				}
			}

			// students should only get moduleitems with a visible activity
			if userIsInstructor(r) || item.Activity.Visible {
				module.Items = append(module.Items, item)
			}
		}

		// students should only get modules with moduleitems with visible activities
		if userIsInstructor(r) || len(module.Items) > 0 {
			modules = append(modules, module)
		}

	}

	writeJSON(w, r, modules)
}

// ModuleInstructionPUT updates the instruction for the given module
func (api *API) ModuleInstructionPUT(w http.ResponseWriter, r *http.Request) {

	// only allow teacher to edit instruction
	if !assertUserIsInstructor(w, r) {
		return
	}

	vars := mux.Vars(r)
	moduleID, _ := strconv.Atoi(vars["id"])

	defer r.Body.Close()
	var instruction StringValue

	if err := json.NewDecoder(r.Body).Decode(&instruction); err != nil {
		http.Error(w, fmt.Sprintf("failed to decode payload: %v", err), http.StatusBadRequest)
		return
	}

	if err := api.dataStore.InstructionSave(moduleID, instruction.Value); err != nil {
		log.Printf("API.ModuleInstructionPUT: %v", err)
		http.Error(w, "failed to update module instruction", http.StatusInternalServerError)
		return
	}

	w.WriteHeader(http.StatusOK)

	// get instruction from database and create response
	var resp string
	i, err := api.dataStore.InstructionByModuleID(moduleID)
	if err != nil {
		if err != datastore.ErrNoResult {
			log.Printf("failed to create response: %v", err)
		}
	} else {
		resp = i.Instruction
	}

	if err := json.NewEncoder(w).Encode(resp); err != nil {
		log.Printf("failed to encode JSON response: %v", err)
	}
}

// ActivityVisiblePUT updates the visibility of activity planned for the given moduleitem
func (api *API) ActivityVisiblePUT(w http.ResponseWriter, r *http.Request) {

	// only allow teacher to edit activity
	if !assertUserIsInstructor(w, r) {
		return
	}

	vars := mux.Vars(r)
	moduleItemID, _ := strconv.Atoi(vars["id"])

	defer r.Body.Close()
	var visible bool

	if err := json.NewDecoder(r.Body).Decode(&visible); err != nil {
		http.Error(w, fmt.Sprintf("failed to decode payload: %v", err), http.StatusBadRequest)
		return
	}

	// get id of course from session
	courseID, ok := r.Context().Value(ContextKey("courseID")).(int)
	if !ok {
		log.Print("API.ActivityVisiblePUT: invalid courseID")
		http.Error(w, http.StatusText(http.StatusInternalServerError), http.StatusInternalServerError)
		return
	}

	// add activity to database (if not already added)
	if err := api.dataStore.ActivityAdd(moduleItemID, courseID); err != nil {
		log.Printf("API.ActivityVisiblePUT: %v", err)
		http.Error(w, "failed to update activity", http.StatusInternalServerError)
		return
	}

	// update visibility of activity
	if err := api.dataStore.ActivityVisibleSave(moduleItemID, visible); err != nil {
		log.Printf("API.ActivityVisiblePUT: %v", err)
		http.Error(w, "failed to update activity", http.StatusInternalServerError)
		return
	}

	w.WriteHeader(http.StatusNoContent)
}

// ActivityStartStopPUT updates start- and stopdate of activity planned for the given moduleitem
func (api *API) ActivityStartStopPUT(w http.ResponseWriter, r *http.Request) {

	// only allow teacher to edit activity
	if !assertUserIsInstructor(w, r) {
		return
	}

	vars := mux.Vars(r)
	moduleItemID, _ := strconv.Atoi(vars["id"])

	defer r.Body.Close()
	var startstop StartStop

	if err := json.NewDecoder(r.Body).Decode(&startstop); err != nil {
		http.Error(w, fmt.Sprintf("failed to decode payload: %v", err), http.StatusBadRequest)
		return
	}

	// get id of course from session
	courseID, ok := r.Context().Value(ContextKey("courseID")).(int)
	if !ok {
		log.Print("API.ActivityStartStopPUT: invalid courseID")
		http.Error(w, http.StatusText(http.StatusInternalServerError), http.StatusInternalServerError)
		return
	}

	// add activity to database (if not already added)
	if err := api.dataStore.ActivityAdd(moduleItemID, courseID); err != nil {
		log.Printf("API.ActivityVisiblePUT: %v", err)
		http.Error(w, "failed to update activity", http.StatusInternalServerError)
		return
	}

	// update start- and stopdate of activity
	if err := api.dataStore.ActivityStartStopSave(moduleItemID, startstop.Start, startstop.Stop); err != nil {
		log.Printf("API.ActivityStartStopPUT: %v", err)
		http.Error(w, "failed to update activity", http.StatusInternalServerError)
		return
	}

	w.WriteHeader(http.StatusNoContent)
}

// ActivityEstimatedPUT updates the activity planned for the given moduleitem
func (api *API) ActivityEstimatedPUT(w http.ResponseWriter, r *http.Request) {

	// only allow teacher to edit activity
	if !assertUserIsInstructor(w, r) {
		return
	}

	vars := mux.Vars(r)
	moduleItemID, _ := strconv.Atoi(vars["id"])

	defer r.Body.Close()
	var estimated FloatValue

	if err := json.NewDecoder(r.Body).Decode(&estimated); err != nil {
		http.Error(w, fmt.Sprintf("failed to decode payload: %v", err), http.StatusBadRequest)
		return
	}

	// get id of course from session
	courseID, ok := r.Context().Value(ContextKey("courseID")).(int)
	if !ok {
		log.Print("API.ActivityEstimatedPUT: invalid courseID")
		http.Error(w, http.StatusText(http.StatusInternalServerError), http.StatusInternalServerError)
		return
	}

	// add activity to database (if not already added)
	if err := api.dataStore.ActivityAdd(moduleItemID, courseID); err != nil {
		log.Printf("API.ActivityVisiblePUT: %v", err)
		http.Error(w, "failed to update activity", http.StatusInternalServerError)
	}

	// update estimated time of activity
	if err := api.dataStore.ActivityEstimatedSave(moduleItemID, estimated.Value); err != nil {
		log.Printf("API.ActivityEstimatedPUT: %v", err)
		http.Error(w, "failed to update activity", http.StatusInternalServerError)
		return
	}

	w.WriteHeader(http.StatusNoContent)
}

// ActivityCategoriesPOST sets category for activity planned for the given moduleitem
func (api *API) ActivityCategoriesPOST(w http.ResponseWriter, r *http.Request) {

	// only allow teacher to edit activity
	if !assertUserIsInstructor(w, r) {
		return
	}

	vars := mux.Vars(r)
	moduleItemID, _ := strconv.Atoi(vars["id"])

	defer r.Body.Close()
	var category Category

	if err := json.NewDecoder(r.Body).Decode(&category); err != nil {
		http.Error(w, fmt.Sprintf("failed to decode payload: %v", err), http.StatusBadRequest)
		return
	}

	// get id of course from session
	courseID, ok := r.Context().Value(ContextKey("courseID")).(int)
	if !ok {
		log.Print("API.ActivityEstimatedPUT: invalid courseID")
		http.Error(w, http.StatusText(http.StatusInternalServerError), http.StatusInternalServerError)
		return
	}

	// add activity to database (if not already added)
	if err := api.dataStore.ActivityAdd(moduleItemID, courseID); err != nil {
		log.Printf("API.ActivityVisiblePUT: %v", err)
		http.Error(w, "failed to update activity", http.StatusInternalServerError)
	}

	// add category to activity
	if err := api.dataStore.ActivityCategoryAdd(moduleItemID, category.ID); err != nil {
		log.Printf("API.ActivityCategoriesPOST: %v", err)
		http.Error(w, "failed to add category", http.StatusInternalServerError)
		return
	}

	w.WriteHeader(http.StatusNoContent)
}

// ActivityCategoryDELETE removes category for activity planned for the given moduleitem
func (api *API) ActivityCategoryDELETE(w http.ResponseWriter, r *http.Request) {

	// only allow teacher to edit activity
	if !assertUserIsInstructor(w, r) {
		return
	}

	vars := mux.Vars(r)
	moduleItemID, _ := strconv.Atoi(vars["id"])
	categoryID, _ := strconv.Atoi(vars["category_id"])

	if err := api.dataStore.ActivityCategoryRemove(moduleItemID, categoryID); err != nil {
		log.Printf("API.ActivityCategoryDELETE: %v", err)
		http.Error(w, "failed to remove category", http.StatusInternalServerError)
		return
	}

	w.WriteHeader(http.StatusNoContent)
}

// ActivityUserEstimatedPUT updates estimated time of the ctivity planned for the given moduleitem by current user
func (api *API) ActivityUserEstimatedPUT(w http.ResponseWriter, r *http.Request) {

	vars := mux.Vars(r)
	moduleItemID, _ := strconv.Atoi(vars["id"])

	defer r.Body.Close()
	var estimated FloatValue

	if err := json.NewDecoder(r.Body).Decode(&estimated); err != nil {
		http.Error(w, fmt.Sprintf("failed to decode payload: %v", err), http.StatusBadRequest)
		return
	}

	// get id of user from session
	userID, ok := r.Context().Value(ContextKey("userID")).(int)
	if !ok {
		log.Print("API.ActivityUserEstimatedPUT: invalid userID")
		http.Error(w, http.StatusText(http.StatusInternalServerError), http.StatusInternalServerError)
		return
	}

	if err := api.dataStore.ActivityUserEstimatedSave(moduleItemID, userID, estimated.Value); err != nil {
		log.Printf("API.ActivityUserEstimatedPUT: %v", err)
		http.Error(w, "failed to update activity", http.StatusInternalServerError)
		return
	}

	w.WriteHeader(http.StatusNoContent)

}

// CalendarEventPOST imports activity planned for the given moduleitem to calendar of current user
func (api *API) CalendarEventPOST(w http.ResponseWriter, r *http.Request) {

	vars := mux.Vars(r)
	moduleItemID, _ := strconv.Atoi(vars["id"])

	defer r.Body.Close()
	var event CalendarEvent

	if err := json.NewDecoder(r.Body).Decode(&event); err != nil {
		http.Error(w, fmt.Sprintf("failed to decode payload: %v", err), http.StatusBadRequest)
		return
	}

	// get modules and module items from canvas
	canvasClient, ok := r.Context().Value(ContextKey("canvasClient")).(*canvas.Client)
	if !ok {
		log.Print("API.Modules: no Canvas client in context")
		http.Error(w, http.StatusText(http.StatusInternalServerError), http.StatusInternalServerError)
		return
	}

	// Post values of event to canvas api
	// if ok do below if not http.Error

	w.WriteHeader(http.StatusOK)

	// get id of user from session
	userID, ok := r.Context().Value(ContextKey("userID")).(int)
	if !ok {
		log.Print("API.CalendarEventPOST: invalid userID")
		http.Error(w, http.StatusText(http.StatusInternalServerError), http.StatusInternalServerError)
		return
	}
	// Call funtion posting to canvas calendar
	if _, err := canvasClient.CalendarEventPost(moduleItemID, userID, event.StartDate, event.StopDate, event.Title, event.Description); err != nil {
		log.Printf("API.ActivityUserPOST: %v", err)
		http.Error(w, "failed to update activity", http.StatusInternalServerError)
		return
	}

	// End post
	w.WriteHeader(http.StatusOK)

	// log import to database
	if err := api.dataStore.CalendarImportSave(moduleItemID, userID); err != nil {
		log.Printf("failed to log calendar import: %v", err)
	}

	// create response with last import date
	au, err := api.dataStore.ActivityUserByModuleItemID(moduleItemID, userID)
	if err != nil {
		log.Printf("failed to create response: %v", err)
	}

	if err := json.NewEncoder(w).Encode(au.Imported); err != nil {
		log.Printf("failed to encode JSON response: %v", err)
	}
}

// Planning returns planning for current course
func (api *API) Planning(w http.ResponseWriter, r *http.Request) {

	// get id of current course from session
	courseID, ok := r.Context().Value(ContextKey("courseID")).(int)
	if !ok {
		log.Print("API.Planning: invalid courseID")
		http.Error(w, http.StatusText(http.StatusInternalServerError), http.StatusInternalServerError)
		return
	}

	// get planning for current course from db
	planning := Planning{}
	p, err := api.dataStore.PlanningByCourseID(courseID)
	if err != nil {
		if err != datastore.ErrNoResult {
			log.Printf("API.Planning: %v", err)
			http.Error(w, "failed to load planning for course", http.StatusInternalServerError)
			return
		}
	}
	planning.Activated = p.Activated

	writeJSON(w, r, planning)
}

// PlanningActivePUT sets activated status of planning for current course
func (api *API) PlanningActivePUT(w http.ResponseWriter, r *http.Request) {

	defer r.Body.Close()
	var active bool

	if err := json.NewDecoder(r.Body).Decode(&active); err != nil {
		http.Error(w, fmt.Sprintf("failed to decode payload: %v", err), http.StatusBadRequest)
		return
	}

	// get id of course from session
	courseID, ok := r.Context().Value(ContextKey("courseID")).(int)
	if !ok {
		log.Print("API.PlanningActivePUT: invalid courseID")
		http.Error(w, http.StatusText(http.StatusInternalServerError), http.StatusInternalServerError)
		return
	}

	if err := api.dataStore.PlanningActiveSave(courseID, active); err != nil {
		log.Printf("API.PlanningActivePUT: %v", err)
		http.Error(w, "failed to update planning", http.StatusInternalServerError)
		return
	}

	w.WriteHeader(http.StatusOK)

	// get activated date from database and create response
	var resp *time.Time
	p, err := api.dataStore.PlanningByCourseID(courseID)
	if err != nil {
		if err != datastore.ErrNoResult {
			log.Printf("failed to create response: %v", err)
		}
	} else {
		resp = p.Activated
	}

	if err := json.NewEncoder(w).Encode(resp); err != nil {
		log.Printf("failed to encode JSON response: %v", err)
	}
}

// Categories returns available categories for planned activities
func (api *API) Categories(w http.ResponseWriter, r *http.Request) {
	var categories []Category

	dbCategories, err := api.dataStore.Categories()

	if err != nil {
		if err != datastore.ErrNoResult {
			log.Printf("API.Categories: %v", err)
			http.Error(w, "failed to load categories", http.StatusInternalServerError)
			return
		}
	}

	for _, c := range dbCategories {
		categories = append(categories, Category{
			ID:   c.ID,
			Name: c.Name,
		})
	}

	writeJSON(w, r, categories)
}
