package datastore

import (
	"database/sql"
	"errors"
	"fmt"
	"time"

	_ "github.com/go-sql-driver/mysql" // MySQL
)

// Datastore represents a datastore
type Datastore struct {
	db *sql.DB
}

// New returns a new datastore
func New(urn string) (*Datastore, error) {
	db, err := sql.Open("mysql", urn)
	if err != nil {
		return nil, fmt.Errorf("failed to open database: %v", err)
	}
	if err = db.Ping(); err != nil {
		return nil, err
	}
	ds := &Datastore{
		db: db,
	}
	return ds, nil
}

// Token represents an OAuth2 token
type Token struct {
	AccessToken  string
	RefreshToken string
	TokenType    string
	Expiry       time.Time
}

// Category represents a category that can be set for a planned activity
type Category struct {
	ID   int
	Name string
}

// Instruction represents an instruction for a module
type Instruction struct {
	ID          int
	ModuleID    int
	Instruction string
}

// Activity represents a planned activity for a module item
type Activity struct {
	ID           int
	ModuleItemID int
	StartDate    *time.Time
	StopDate     *time.Time
	Estimated    *float32
	Updated      time.Time
	Visible      bool
	CourseID     int
	Categories   []Category
}

// ActivityUser represents a planned activity for a module item for a specific user
type ActivityUser struct {
	ID           int
	ModuleItemID int
	UserID       int
	Estimated    *float32
	Imported     *time.Time
}

// Planning represents the planning for a specific course
type Planning struct {
	ID        int
	CourseID  int
	Activated *time.Time
}

// ErrNoResult is used when a query returns no result. Corresponds to sql.ErrNoRows.
var ErrNoResult = errors.New("no result found")

// Token retrieves a token
func (ds *Datastore) Token(userID string) (Token, error) {
	var t Token

	row := ds.db.QueryRow("SELECT access,refresh,tokentype,expiry FROM tokens WHERE userid=?;", userID)
	if err := row.Scan(&t.AccessToken, &t.RefreshToken, &t.TokenType, &t.Expiry); err != nil {
		if err == sql.ErrNoRows {
			return t, ErrNoResult
		}
		return t, fmt.Errorf("failed to fetch token: %v", err)
	}

	return t, nil
}

// TokenSave saves a token
func (ds *Datastore) TokenSave(userid, accessToken, refreshToken, tokenType string, expiry time.Time) error {
	_, err := ds.db.Exec("INSERT INTO tokens (userid,access,refresh,tokentype,expiry) VALUES(?,?,?,?,?) ON DUPLICATE KEY UPDATE access=VALUES(access), expiry=VALUES(expiry)", userid, accessToken, refreshToken, tokenType, expiry)
	if err != nil {
		return fmt.Errorf("failed to save token in database: %v", err)
	}

	return nil
}

// ActivityByModuleItemID gets activity planned for the given module item
func (ds *Datastore) ActivityByModuleItemID(moduleItemID int) (Activity, error) {
	query := "SELECT id, moduleitemid, startdate, stopdate, estimated, updated, visible, courseid" +
		" FROM activity" +
		" WHERE moduleitemid=?;"
	row := ds.db.QueryRow(query, moduleItemID)

	a := Activity{}
	if err := row.Scan(&a.ID, &a.ModuleItemID, &a.StartDate, &a.StopDate, &a.Estimated, &a.Updated, &a.Visible, &a.CourseID); err != nil {
		if err == sql.ErrNoRows {
			return a, ErrNoResult
		}
		return a, fmt.Errorf("failed to fetch activity for module item %d: %v", moduleItemID, err)
	}

	c, err := ds.CategoriesByActivityID(a.ID)
	if err != nil {
		return a, fmt.Errorf("failed to fetch categories for module item %d: %v", moduleItemID, err)
	}
	a.Categories = c

	return a, nil
}

// InstructionByModuleID gets instruction for the given module
func (ds *Datastore) InstructionByModuleID(moduleID int) (Instruction, error) {
	query := "SELECT id, moduleid, instruction" +
		" FROM instruction" +
		" WHERE moduleid=?;"
	row := ds.db.QueryRow(query, moduleID)

	i := Instruction{}
	if err := row.Scan(&i.ID, &i.ModuleID, &i.Instruction); err != nil {
		if err == sql.ErrNoRows {
			return i, ErrNoResult
		}
		return i, fmt.Errorf("failed to fetch instruction for module %d: %v", moduleID, err)
	}

	return i, nil
}

// InstructionSave sets instruction for the given module
func (ds *Datastore) InstructionSave(moduleID int, instruction *string) error {
	query := "INSERT INTO instruction (moduleid, instruction)" +
		" VALUES(?, ?)" +
		" ON DUPLICATE KEY UPDATE" +
		" instruction=VALUES(instruction);"
	_, err := ds.db.Exec(query, moduleID, instruction)
	if err != nil {
		return fmt.Errorf("failed to save instruction for module %d: %v", moduleID, err)
	}

	return nil
}

// CategoriesByActivityID returns all categories set for the given activity
func (ds *Datastore) CategoriesByActivityID(activityID int) ([]Category, error) {
	var categories []Category

	query := "SELECT c.id, c.name" +
		" FROM category c JOIN activity_category ac ON c.id = ac.categoryid" +
		" WHERE ac.activityid=?;"

	rows, err := ds.db.Query(query, activityID)
	if err != nil {
		if err == sql.ErrNoRows {
			return categories, ErrNoResult
		}
		return categories, fmt.Errorf("failed to fetch categories for activity %d: %v", activityID, err)
	}
	defer rows.Close()

	for rows.Next() {
		c := Category{}
		if err := rows.Scan(&c.ID, &c.Name); err != nil {
			return categories, fmt.Errorf("failed to read categories for activity %d: %v", activityID, err)
		}
		categories = append(categories, c)
	}

	return categories, nil
}

// ActivityAdd adds activity planned for given moduleitem (if not already added)
func (ds *Datastore) ActivityAdd(moduleItemID int, courseID int) error {
	query := "INSERT INTO activity (moduleitemid, courseid)" +
		" VALUES(?, ?)" +
		" ON DUPLICATE KEY UPDATE" +
		" courseid=VALUES(courseid);"
	_, err := ds.db.Exec(query, moduleItemID, courseID)
	if err != nil {
		return fmt.Errorf("failed to set activity for moduleitem %d: %v", moduleItemID, err)
	}

	return nil
}

// ActivityVisibleSave sets the visibility of activity planned for the given moduleitem
func (ds *Datastore) ActivityVisibleSave(moduleItemID int, visible bool) error {
	query := "UPDATE activity" +
		" SET visible = ?" +
		" WHERE moduleitemid = ?;"
	_, err := ds.db.Exec(query, visible, moduleItemID)
	if err != nil {
		return fmt.Errorf("failed to set activity visible for moduleitem %d: %v", moduleItemID, err)
	}

	return nil
}

// ActivityStartStopSave sets the start- and stopdates of activity planned for the given moduleitem
func (ds *Datastore) ActivityStartStopSave(moduleItemID int, startdate *time.Time, stopdate *time.Time) error {
	query := "UPDATE activity" +
		" SET startdate = ?," +
		" stopdate = ?," +
		" updated = NOW()" +
		" WHERE moduleitemid = ?;"
	_, err := ds.db.Exec(query, startdate, stopdate, moduleItemID)
	if err != nil {
		return fmt.Errorf("failed to set activity startstop for moduleitem %d: %v", moduleItemID, err)
	}

	return nil
}

// ActivityEstimatedSave sets the estimated time for activity planned for the given moduleitem
func (ds *Datastore) ActivityEstimatedSave(moduleItemID int, estimated *float32) error {
	query := "UPDATE activity" +
		" SET estimated = ?," +
		" updated = NOW()" +
		" WHERE moduleitemid = ?;"
	_, err := ds.db.Exec(query, estimated, moduleItemID)
	if err != nil {
		return fmt.Errorf("failed to set activity estimated for moduleitem %d: %v", moduleItemID, err)
	}

	return nil
}

// ActivityCategoryAdd adds category to the activity planned for the given moduleitem
func (ds *Datastore) ActivityCategoryAdd(moduleItemID int, categoryID int) error {
	tx, err := ds.db.Begin()
	if err != nil {
		return fmt.Errorf("failed to start transaction: %v", err)
	}

	// add category to activity
	query := "INSERT INTO activity_category (activityid, categoryid)" +
		" SELECT id, ?" +
		" FROM activity" +
		" WHERE moduleitemid = ?;"

	_, err = tx.Exec(query, categoryID, moduleItemID)
	if err != nil {
		tx.Rollback()
		return fmt.Errorf("failed to add category %d to activity for moduleitem %d: %v", categoryID, moduleItemID, err)
	}

	// set activity updated
	query = "UPDATE activity" +
		" SET updated = NOW()" +
		" WHERE moduleitemid = ?;"

	_, err = tx.Exec(query, moduleItemID)
	if err != nil {
		tx.Rollback()
		return fmt.Errorf("failed to set updated on add category %d to activity for moduleitem %d: %v", categoryID, moduleItemID, err)
	}

	tx.Commit()

	return nil
}

// ActivityCategoryRemove removes category from the activity planned for the given moduleitem
func (ds *Datastore) ActivityCategoryRemove(moduleItemID int, categoryID int) error {
	tx, err := ds.db.Begin()
	if err != nil {
		return fmt.Errorf("failed to start transaction: %v", err)
	}

	// remove category from activity
	query := "DELETE ac" +
		" FROM activity_category ac" +
		" JOIN activity a ON ac.activityid = a.id" +
		" WHERE a.moduleitemid = ? AND ac.categoryid = ?;"

	_, err = tx.Exec(query, moduleItemID, categoryID)
	if err != nil {
		tx.Rollback()
		return fmt.Errorf("failed to remove category %d for activity for moduleitem %d: %v", categoryID, moduleItemID, err)
	}

	// set activity updated
	query = "UPDATE activity" +
		" SET updated = NOW()" +
		" WHERE moduleitemid = ?;"

	_, err = tx.Exec(query, moduleItemID)
	if err != nil {
		tx.Rollback()
		return fmt.Errorf("failed to set updated on remove category %d for activity for moduleitem %d: %v", categoryID, moduleItemID, err)
	}

	tx.Commit()

	return nil
}

// ActivityUserByModuleItemID gets activity planned for the given module item by a specific user
func (ds *Datastore) ActivityUserByModuleItemID(moduleItemID int, userID int) (ActivityUser, error) {
	query := "SELECT id, moduleitemid, userid, estimated, imported" +
		" FROM activity_user" +
		" WHERE moduleitemid=? AND userid=?;"
	row := ds.db.QueryRow(query, moduleItemID, userID)

	a := ActivityUser{}
	if err := row.Scan(&a.ID, &a.ModuleItemID, &a.UserID, &a.Estimated, &a.Imported); err != nil {
		if err == sql.ErrNoRows {
			return a, ErrNoResult
		}
		return a, fmt.Errorf("failed to fetch activity for module item %d user id %d: %v", moduleItemID, userID, err)
	}

	return a, nil
}

// ActivityUserEstimatedSave sets the estimated time for activity planned for the given moduleitem by a specific user
func (ds *Datastore) ActivityUserEstimatedSave(moduleItemID int, userID int, estimated *float32) error {

	query := "INSERT INTO activity_user (moduleitemid, userid, estimated)" +
		" VALUES(?,?,?)" +
		" ON DUPLICATE KEY UPDATE" +
		" estimated=VALUES(estimated);"

	_, err := ds.db.Exec(query, moduleItemID, userID, estimated)
	if err != nil {
		return fmt.Errorf("failed to set activity for moduleitem %d user %d: %v", moduleItemID, userID, err)
	}

	return nil
}

// CalendarImportSave logs a calendar import for a given moduleitem by a specific user
func (ds *Datastore) CalendarImportSave(moduleItemID int, userID int) error {
	query := "INSERT INTO activity_user (moduleitemid, userid, imported)" +
		" VALUES(?,?, NOW())" +
		" ON DUPLICATE KEY UPDATE" +
		" imported=VALUES(imported);"

	_, err := ds.db.Exec(query, moduleItemID, userID)
	if err != nil {
		return fmt.Errorf("failed to save calendar import for moduleitem %d user %d: %v", moduleItemID, userID, err)
	}

	return nil
}

// PlanningByCourseID gets planning for a given course
func (ds *Datastore) PlanningByCourseID(courseID int) (Planning, error) {

	query := "SELECT id, courseid, activated" +
		" FROM planning" +
		" WHERE courseid=?;"
	row := ds.db.QueryRow(query, courseID)

	p := Planning{}
	if err := row.Scan(&p.ID, &p.CourseID, &p.Activated); err != nil {
		if err == sql.ErrNoRows {
			return p, ErrNoResult
		}
		return p, fmt.Errorf("failed to fetch planning for course %d: %v", courseID, err)
	}

	return p, nil
}

// PlanningActiveSave sets activated status of planning
func (ds *Datastore) PlanningActiveSave(courseID int, active bool) error {
	query := "INSERT INTO planning (courseid, activated)" +
		" VALUES(?, CASE WHEN ? THEN NOW() ELSE NULL END)" +
		" ON DUPLICATE KEY UPDATE" +
		" activated=VALUES(activated);"
	_, err := ds.db.Exec(query, courseID, active)
	if err != nil {
		return fmt.Errorf("failed to set planning activated for course %d: %v", courseID, err)
	}

	return nil
}

// Categories returns all available categories for planned activities
func (ds *Datastore) Categories() ([]Category, error) {
	var categories []Category

	rows, err := ds.db.Query("SELECT id, name FROM category;")
	if err != nil {
		if err == sql.ErrNoRows {
			return categories, ErrNoResult
		}
		return categories, fmt.Errorf("failed to fetch categories: %v", err)
	}
	defer rows.Close()

	for rows.Next() {
		c := Category{}
		if err := rows.Scan(&c.ID, &c.Name); err != nil {
			return categories, fmt.Errorf("failed to read categories: %v", err)
		}
		categories = append(categories, c)
	}

	return categories, nil
}
